unit RTFRPT;

{ RTF Report (C)2K by Paul TOTH <tothpaul@multimania.com>
  http://www.multimania.com/tothpaul
  http://tothpaul.free.fr
}

{
  Need a fast way to create Word Documents ?
  Do what Microsoft Access do, create a RTF document :)
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TMetrics=(mtMM,mtTwips);
  TTabStyle=(tsLeft,tsRight,tsCenter,tsDecimal);
  TTextAlignment=(taLeft,taRight,taCenter,taJustified);
  TBorderStyle=(bsSingle,bsDouble,bsShadow,bsDotted,bsDashed,bsEmboss);
  TBorderPos=(bpTop,bpLeft,bpRight,bpBottom,bpBox);

  TSection=record
   Font:TFont;
   Text:TStringList;
  end;

  TRTFReport = class(TComponent)
  private
   fRTF        :TStringList;        // RTF doc
   fFonts      :TStringList;        // List of fonts for current doc
   fSections   :TList;              // List of Sections for current doc
   fSection    :TStringList;        // Current section text
   fFont       :TFont;              // Current Font
   fOldFont    :TFont;              // Previous Font (to compare with)
   fStartFont  :TFont;              // Default font
   fMetrics    :TMetrics;           // choice between mm and Twips
   fPageHeight :integer;
   fPageWidth  :integer;
   fMargeLeft  :integer;
   fMargeTop   :integer;
   fMargeRight :integer;
   fMargeBottom:integer;
   fIdentFirst :integer;
   fIdentLeft  :integer;
   fIdentRight :integer;
   EOnPrint    :TNotifyEvent;       // Dummy Events :
   EOnTitle    :TNotifyEvent;       //  use them as you want :)
   EOnHeaders  :TNotifyEvent;
   EOnDetail   :TNotifyEvent;
   EOnPage     :TNotifyEvent;
   procedure SetMetrics(Value:TMetrics);
   procedure SetIdentFirst(Value:integer);
   procedure SetIdentLeft(Value:integer);
   procedure SetIdentRight(Value:integer);
   Procedure SetFont(Value:TFont);
   Procedure FontChange(Sender:TObject);
  protected
   function Twips(Value:integer):string;
   procedure Add(s:string);
   procedure CloseSection;
  public
   Constructor Create(AOwner:TComponent); override;
   Destructor Destroy; override;

   Procedure BeginDoc; // Start a new RTF doc
   Procedure EndDoc;   // end of doc

   Procedure BeginSection; // begin a section
   Procedure EndSection;   // end the section (restore font,...)

   Procedure Align(Alignment:TTextAlignment);
   Procedure AddTabs(Style:TTabStyle; Tabs:array of integer);
   Procedure AddBorder(Where:TBorderPos;Style:TBorderStyle);

   Procedure Write(s:string);
   Procedure WriteLn(s:string);

   Procedure Print;   // Call OnPrint -or- Call Title,Header,Detail and Page
   Procedure Title;   // Call OnTitle
   Procedure Headers; // Call OnHeaders
   Procedure Detail;  // Call OnDetail
   Procedure Page;    // Call OnPage

   Property RTF:TStringList read fRTF;
  published

   Property Metrics:TMetrics    read fMetrics     write SetMetrics   default mtMM;

   Property PageHeight:integer  read fPageHeight  write fPageHeight  default 297;
   Property Pagewidth:integer   read fPageWidth   write fPageWidth   default 210;
   Property MargeLeft:integer   read fMargeLeft   write fMargeLeft   default 25;
   Property MargeTop:integer    read fMargeTop    write fMargeTop    default 25;
   Property MargeRight:integer  read fMargeRight  write fMargeRight  default 25;
   Property MargeBottom:integer read fMargeBottom write fMargeBottom default 25;

   Property IdentFirst:integer  read fIdentFirst  write SetIdentFirst default 10;
   Property IdentLeft :integer  read fIdentLeft   write SetIdentLeft  default 10;
   Property IdentRight:integer  read fIdentRight  write SetIdentRight default 10;

   Property Font:TFont read fFont write SetFont;

   Property OnPrint  :TNotifyEvent read EOnPrint   write EOnPrint;
   Property OnTitle  :TNotifyEvent read EOnTitle   write EOnTitle;
   Property OnHeaders:TNotifyEvent read EOnHeaders write EOnHeaders;
   Property OnDetail :TNotifyEvent read EOnDetail  write EOnDetail;
   Property OnPage   :TNotifyEvent read EOnPage    write EOnPage;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MySoft.Report', [TRTFReport]);
end;

Constructor TRTFReport.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fRTF:=TStringList.Create;
  fSections:=TList.Create;
  fSection:=nil;
  fFonts:=TStringList.Create;
  fFont:=TFont.Create;
  fFont.OnChange:=FontChange;
  fOldFont:=TFont.Create;
  fStartFont:=TFont.Create;
  fMetrics:=mtMM;
  fPageHeight:=297; // A4 = 21 x 29,7
  fPageWidth:=210;
  fMargeLeft:=25;   // 2,5 cm
  fMargeRight:=25;
  fMargeTop:=25;
  fMargeBottom:=25;
  fIdentFirst:=10;  // 1 cm
  fIdentLeft:=0;
  fIdentRight:=0;
 end;

Destructor TRTFReport.Destroy;
 begin
  fStartFont.Free;
  fOldFont.Free;
  fFont.Free;
  fFonts.Free;
  fSection.Free;
  fSections.Free;
  fRTF.Free;
  Inherited Destroy;
 end;

function TRTFReport.Twips(Value:integer):string;
 begin
  case fMetrics of
   mtMM : Value:=(Value*567+5) div 10;
  end;
  Result:=IntToStr(Value);
 end;

procedure TRTFReport.Add(s:string);
 var
  i:integer;
 begin
  if fSection=nil then BeginSection;
  i:=fSection.Count-1;
  if i<0 then
   fSection.Add(s)
  else
   fSection[i]:=fSection[i]+s;
 end;

Procedure TRTFReport.SetMetrics(Value:TMetrics);
 Const
  K:array[TMetrics] of integer=(10,567);
 var
  m,d,a:integer;
  procedure MulDiv(Var i:integer);
   begin
    i:=(i*m+a) div d;
   end;
 begin
  if Value<>fMetrics then begin
   m:=K[Value];
   d:=K[fMetrics];
   a:=d div 2;
   MulDiv(fPageWidth);
   MulDiv(fPageHeight);
   MulDiv(fMargeLeft);
   MulDiv(fMargeTop);
   MulDiv(fMargeRight);
   MulDiv(fMargeBottom);
   MulDiv(fIdentFirst);
   MulDiv(fIdentLeft);
   MulDiv(fIdentRight);
   fMetrics:=Value;
  end;
 end;

Procedure TRTFReport.AddBorder(Where:TBorderPos;Style:TBorderStyle);
 begin
  case Where of
   bpBox    : Add('\brdrx');
   bpTop    : Add('\brdrt');
   bpLeft   : Add('\brdrl');
   bpRight  : Add('\brdrr');
   bpBottom : Add('\brdrb');
  end;
  case Style of
   bsSingle: Add('\brdrs');
   bsDouble: Add('\brdrdb');
   bsShadow: Add('\brdrsh');
   bsDotted: Add('\brdrdot');
   bsDashed: Add('\brdrdash');
   bsEmboss: Add('\brdremboss');
  end;
 end;

Procedure TRTFReport.SetFont(Value:TFont);
 begin
  fFont.Assign(Value);
 end;

Procedure TRTFReport.FontChange(Sender:TObject);
 var
  s:string;
  i:integer;

  procedure SetStyle(Style:TFontStyle; OnOff:string);
   begin
    if Style in fFont.Style then begin
     if not (Style in fOldFont.Style) then s:=s+OnOff;
    end else begin
     if (Style in fOldFont.Style) then s:=s+OnOff+'0';
    end;
   end;

 begin
  s:=fFont.Name;
  if s=fOldFont.Name then
   s:=''
  else begin
   i:=fFonts.IndexOf(s);
   if i<0 then i:=fFonts.Add(s);
   s:='\f'+IntToStr(i+1);
  end;
  if fFont.Size<>fOldFont.Size then s:=s+'\fs'+IntToStr(2*fFont.Size);
  SetStyle(fsItalic,'\i');
  SetStyle(fsBold,'\b');
  SetStyle(fsUnderLine,'\ul');
  SetStyle(fsStrikeOut,'\strike');
  if s<>'' then Add(s);
  fOldFont.Assign(fFont);
 end;

Procedure TRTFReport.BeginDoc;
 var
  s:string;
 begin
  fStartFont.Assign(fFont);
  fRTF.Clear;
  BeginSection;
  Add('\rtf1\ansi\deff0');
  Write('\paperw'+Twips(fPageWidth)+
        '\paperh'+Twips(fPageHeight)+
        '\margl' +Twips(fMargeLeft)+
        '\margr' +Twips(fMargeRight)+
        '\margt' +Twips(fMargeTop)+
        '\margb' +Twips(fMargeBottom));
  if fIdentFirst<>0 then Add('\fi'+Twips(fIdentFirst));
  if fIdentLeft <>0 then Add('\li'+Twips(fIdentLeft));
  if fIdentRight<>0 then Add('\ri'+Twips(fIdentRight));
  fFonts.Clear;
  fOldFont.Assign(fFont);
  fFonts.Add(fFont.Name);
  BeginSection;
  s:='\f1\fs'+IntToStr(2*fFont.Size); // Some fonts want work on f0 ?!!
  if fsItalic    in fFont.Style then s:=s+'\i';
  if fsBold      in fFont.Style then s:=s+'\b';
  if fsUnderLine in fFont.Style then s:=s+'\ul';
  if fsStrikeOut in fFont.Style then s:=s+'\strike';
  Add(s);
 end;

Procedure TRTFReport.EndDoc;
 var
  i,j:integer;
  s:string;
 begin
  if fSections.Count=0 then exit;
  fSection:=TSection(fSections[0]^).Text;
  if fFonts.Count>0 then begin
   Write('{\fonttbl');
   for i:=0 to fFonts.Count-1 do begin
//  \fnil | \froman | \fswiss | \fmodern | \fscript | \fdecor | \ftech | \fbidi
    Write(' {\f'+IntToStr(i+1)+'\fnil '+fFonts[i]+';}');
   end;
   Write('}');
  end;
  fSection:=TSection(fSections.Last^).Text;
  while fSections.Count>0 do CloseSection;
  fOldFont.Assign(fStartFont);
  fFont.Assign(fStartFont);
 end;


Procedure TRTFReport.BeginSection;
 var
  S:^TSection;
 begin
  New(S);
  S^.Font:=TFont.Create;
  S^.Font.Assign(Font);
  S^.Text:=TStringList.Create;
  fSections.Add(S);
  fSection:=S.Text;
  Write('{');
 end;

Procedure TRTFReport.EndSection;
 begin
  if fSections.Count>1 then CloseSection;
 end;

Procedure TRTFReport.CloseSection;
 var
  S:^TSection;
 begin
  if fSections.Count=0 then exit;
  Write('}');
  S:=fSections.Last;
  fSections.Remove(S);
  if fSections.Count=0 then begin
   fSection:=nil;
   fRTF.Assign(S.Text);
  end else begin
   fSection:=TSection(fSections.Last^).Text;
   fSection.AddStrings(S^.Text);
  end;
  fOldFont.Assign(S.Font);
  fFont.Assign(S.Font);
  S.Font.Free;
  S.Text.Free;
  Dispose(S);
 end;

Procedure TRTFReport.SetIdentFirst(Value:integer);
 begin
  if fIdentFirst<>Value then begin
   fIdentFirst:=Value;
   Add('\fi'+Twips(fIdentFirst));
  end;
 end;

Procedure TRTFReport.SetIdentLeft(Value:integer);
 begin
  if fIdentLeft<>Value then begin
   fIdentLeft:=Value;
   Add('\li'+Twips(fIdentLeft));
  end;
 end;

Procedure TRTFReport.SetIdentRight(Value:integer);
 begin
  if fIdentRight<>Value then begin
   fIdentRight:=Value;
   Add('\ri'+Twips(fIdentRight));
  end;
 end;

Procedure TRTFReport.Align(Alignment:TTextAlignment);
 const
  Aligns:array[TTextAlignment] of char='lrcj';
 begin
  Add('\q'+Aligns[Alignment]);
 end;

Procedure TRTFReport.AddTabs(Style:TTabStyle; Tabs:array of integer);
 var
  i:integer;
  s,ss:string;
 begin
  case Style of
   tsLeft   : s:='';
   tsRight  : s:='\tqr';
   tsCenter : s:='\tqc';
   tsDecimal: s:='\tqdec';
  end;
  ss:='';
  for i:=low(Tabs) to High(Tabs) do begin
   ss:=ss+s+'\tx'+Twips(Tabs[i]);
  end;
  Write(ss);
 end;

Function NoTabs(s:string):string;
 var
  i:integer;
 begin
  i:=pos(#9,s);
  while i<>0 do begin
   Delete(s,i,1);
   Insert('\tab ',s,i);
   i:=pos(#9,s);
  end;
  Result:=s;
 end;

Procedure TRTFReport.Write(s:string);
 begin
  if fSection=nil then BeginSection;
  fSection.Add(NoTabs(s));
 end;

Procedure TRTFReport.WriteLn(s:string);
 begin
  if fSection=nil then BeginSection;
  fSection.Add(NoTabs(s)+'\par');
 end;

Procedure TRTFReport.Print;
 begin;
  if Assigned(EOnPrint) then
   EOnPrint(Self)
  else begin
   BeginDoc;
   Title;
   Headers;
   Detail;
   Page;
   EndDoc;
  end;
 end;

Procedure TRTFReport.Title;
 begin
  if Assigned(EOnTitle) then EOnTitle(Self);
 end;

Procedure TRTFReport.Headers;
 begin
  if Assigned(EOnHeaders) then EOnHeaders(Self);
 end;

Procedure TRTFReport.Detail;
 begin
  if Assigned(EOnDetail) then EOnDetail(Self);
 end;

Procedure TRTFReport.Page;
 begin
  if Assigned(EOnPage) then EOnPage(Self);
 end;


end.

